"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ProtoFloatSeriesValuesDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOTAL_ITEM_COUNT_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    total_item_count: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProtoFloatPointValueDTO]: ...
    def __init__(
        self,
        *,
        total_item_count: builtins.int = ...,
        values: collections.abc.Iterable[global___ProtoFloatPointValueDTO] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["total_item_count", b"total_item_count", "values", b"values"]) -> None: ...

global___ProtoFloatSeriesValuesDTO = ProtoFloatSeriesValuesDTO

@typing_extensions.final
class ProtoFloatPointValueDTO(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_MILLIS_FIELD_NUMBER: builtins.int
    STEP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp_millis: builtins.int
    step: builtins.float
    value: builtins.float
    def __init__(
        self,
        *,
        timestamp_millis: builtins.int = ...,
        step: builtins.float = ...,
        value: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["step", b"step", "timestamp_millis", b"timestamp_millis", "value", b"value"]) -> None: ...

global___ProtoFloatPointValueDTO = ProtoFloatPointValueDTO
